/*
 *  stereo -- Stereo matching software
 *  $Id: delaunaycv.c 160 2004-07-02 12:11:09Z sam $
 *  Copyright (c) 2004 Sam Hocevar <sam@zoy.org>
 *                     Éric Beets <ericbeets@free.fr>
 *  All Rights Reserved
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307  USA
 */

#include <stdint.h>

#include <cv.h>
#include <stdio.h>

#include "defs.h"
#include "image.h"
#include "main.h"

#include <math.h>

#include <cv.h>
#include <highgui.h>

#include "geometry.h"


#define TRANS_X 41
#define TRANS_Y 3

#define LIMIT_X 21

#define CORR_X 3 
#define CORR_Y 3

#define WINTRANS_Y (CORR_Y+TRANS_Y-1)

/* Continuity weight (16 == normal) */
#define X_WEIGHT 16
#define Y_WEIGHT 16

typedef struct _window Window;

struct _window
{
	int r,g,b;
};

static int Calcul(Window *Refer,Window *Trans,int u,int v);
int Translation(image_t *left, image_t *right,
                image_t *left_depth, image_t *right_depth,
                int p, int q, double adjust,Window *Refer,Window *Trans);


/*
 *     Appariement d'un couple d'image par minimisation de la correlation
 */
int Correlation(image_t *left_norm,
                image_t *right_norm,
                image_t *left_depthmap_norm,
                image_t *right_depthmap_norm, double adjust)
{
    int p,q,r,g,b,i,j;
    float min,max;
    int pixel, progress;
    int corr_min,corr_min_loc;
    int Tab_Stat[20];
    image_t *left,*right;
    
    Window *Refer;
    Window *Trans;
   
    /*Allocation des fenetres de parcours*/
    Refer = malloc(sizeof(Window)*CORR_Y*CORR_X);
    Trans = malloc(sizeof(Window)*WINTRANS_Y*TRANS_X);
    /*Changement d'origine des pointeurs*/
    Refer = Refer+CORR_X*(CORR_Y/2)+CORR_X/2; 
    Trans = Trans+TRANS_X*(WINTRANS_Y/2)+TRANS_X/2; 
    
    /*Allocation des images resultat*/    
    left = malloc(sizeof(image_t));
    right = malloc(sizeof(image_t));
    image_new(left,left_norm->width,left_norm->height,1);
    image_new(right,right_norm->width,right_norm->height,1);
    
    left = left_norm;
    right = right_norm;
    
    printf("Calcul correlation\n");
    /*Initialisation des images resultat*/
    cvFillImage( left_depthmap_norm->_private, CV_RGB(255,0,0));
    cvFillImage( right_depthmap_norm->_private, CV_RGB(255,0,0));
    
    /* Initialisation du tableau des statistiques */
    for(p = 0; p < 20; p++)
        Tab_Stat[p] = 0;

    /* Parcours de l'image */
    for(corr_min = 9999999, progress = 0, q = 1; q < left_norm->height - 2; q++)
    {
        while( q* 100 / left_norm->height >= progress)
        {
            printf("\r %i %% effectués", progress);
            fflush(stdout);
            progress ++;
        }

        for(p = 1; p < left_norm->width - 2; p++)
        {
		/*Remplissage de la fenetre de reference */
		for(i = -CORR_X/2; i < CORR_X-CORR_X/2; i++)
		{
			for(j = -CORR_Y/2; j < CORR_Y-CORR_Y/2; j++)
		       	{
		            getpixel_fast(left, i + p, j + q + adjust, &r, &g, &b);
			    (Refer+j*CORR_X+i)->r = r;
			    (Refer+j*CORR_X+i)->g = g;
			    (Refer+j*CORR_X+i)->b = b;
            		}
            	}
		corr_min_loc = Translation(left, right,
                                       left_depthmap_norm, right_depthmap_norm,
                                       p, q, adjust,Refer,Trans);
		Tab_Stat[corr_min_loc / 1500]++;
        }
    }
    printf("100 %% effectués\n");
    printf("Corr sur l'image %d\n",corr_min);
    
    /*Normalisation de l'affichage*/
    for(min = 999999, max = -999999, p = 1; p < left_norm->width - 2; p++)
    {
        for(q = 1; q < left_norm->height - 2; q++)
        {
            getpixel(left_depthmap_norm, p, q, &r, &g, &b);
            if(r == g)
            {
                if(r < min)
                    min = r;
                if(r < 20 && r > max)
                    max = r;
            }
        }
    }
    printf("min/max: %g %g\n",min,max);
    for(p = 1; p < left_norm->width - 2; p++)
    {
        for(q = 1; q < left_norm->height - 2; q++)
        {
            getpixel(left_depthmap_norm, p, q, &r, &g, &b);
            if(r == g)
            {
                pixel = ((float)(max-r)/(float)(max-min))*255.;
                setpixel(left_depthmap_norm, p, q, pixel, pixel, pixel);
            }

            getpixel(right_depthmap_norm, p, q, &r, &g, &b);
            if(r == g)
            {
                pixel = ((float)(max-r)/(float)(max-min))*255.;
                setpixel(right_depthmap_norm, p, q, pixel, pixel, pixel);
            }
        }
    }
    /*Statistiques de la correlation*/
    printf("Statistiques de correlation par pixel\n");
    for(p=0 ; p<20;p++)
        printf("%d - %d => %d\n",p * 1500, (p+1) * 1500, Tab_Stat[p]);
    
    /*Sauvegarde des images resultats*/
    cvSaveImage("sortie.bmp", left_depthmap_norm->_private);
    cvSaveImage("sortie_.bmp", left_norm->_private);
    return 0;
}
/*
 *     Calcul de la correlation
 */
static inline int Calcul(Window *Refer,Window *Trans,int u,int v)
{
    int i,j;
    int corr = 0;
    int delta;
    Window P1,P2; 
    for(j = -CORR_Y/2 ; j < CORR_Y-CORR_Y/2; j++)
    {
    	int k = j*CORR_X;
	int l = (j+v)*TRANS_X;
	for(i = -CORR_X/2 ; i < CORR_X-CORR_X/2; i++)
        {
	    P1 = Refer[k + i]; 
	    P2 = Trans[l+i+u]; 
	    /*Distance euclidienne*/
            delta = sqrt((P1.r-P2.r)*(P1.r-P2.r)+(P1.g-P2.g)*(P1.g-P2.g)+(P1.b-P2.b)*(P1.b-P2.b));
            corr += delta;
	}
    }
    return corr;
}

/*
 *    Parcours de la fenetre de recherche
 *    Minimisation de la correlation
 */
int Translation(image_t *left, image_t *right,
                image_t *left_depth, image_t *right_depth,
                int p, int q, double adjust_y,Window *Refer,Window *Trans)
{
    int u,v,r,g,b;
    int umin=0,vmin=0;
    double depth;
    int corr;
    int uold_x=0, uold_y=0;
    int good_x=0, good_y=0;
    int corr_min = 99999999;
    /* Premiere colonne */
    if(p == 1)
        uold_x = 0;
    else
    {
        getpixel(left_depth, p - 1, q, &r, &g, &b);
        uold_x = r - LIMIT_X;
        if(r != g || uold_x < - LIMIT_X / 2 || uold_x > LIMIT_X - LIMIT_X / 2)
            uold_x = 0;
        else
            good_x = 1;
    }
    /* Premiere ligne */
    if(q == 1)
        uold_y = 0;
    else
    {
        getpixel(left_depth, p, q - 1, &r, &g, &b);
        uold_y = r - LIMIT_X;
        if(r != g || uold_y < - LIMIT_X / 2 || uold_y > LIMIT_X - LIMIT_X / 2)
            uold_y = 0;
        else
            good_y = 1;
    }
    /* Remplissage de la fenetre de translation */
    for(v = -WINTRANS_Y/2; v < WINTRANS_Y-WINTRANS_Y/2; v++)
    {
	int k = v*TRANS_X;
	for(u = -TRANS_X/2; u < TRANS_X-TRANS_X/2; u++)
        {
	    getpixel_fast(right, p + u + uold_x, q + v, &(Trans+k+u)->r, &(Trans+k+u)->g, &(Trans+k+u)->b);
        }
     }

    /* Parcours de la translation de gauche a droite */
    for(v = -TRANS_Y/2; v < TRANS_Y-TRANS_Y/2; v++)
    {
        for(u = -TRANS_X/2; u < TRANS_X-TRANS_X/2; u++)
        {
	    //corr=16* Calcul_(left,right, p,  q,u,v,adjust_y,uold_x);
            corr = 16 * Calcul(Refer, Trans,u,v);
	    if(good_x)
            {
                corr += X_WEIGHT * (abs(u - uold_x*0));
                if(good_y)
                    corr += Y_WEIGHT * (abs(u - uold_y+uold_x));
            }
            /* Miminisation des correlations de la fenetre de translation */
            if(corr < corr_min)
            {
                corr_min = corr;
                umin = u+uold_x;
                vmin = v;
            }
        }
    }
    depth = umin + LIMIT_X;
    if(corr_min >= 4000)
        setpixel(left_depth, p, q, 255, 0, 0);
    else
        setpixel(left_depth, p, q, depth, depth, depth);
    return corr_min;
}
