/*
 *  stereo -- Stereo matching software
 *  $Id: defs.h 174 2004-07-12 09:39:10Z sam $
 *  Copyright (c) 2004 Sam Hocevar <sam@zoy.org>
 *                     Éric Beets <ericbeets@free.fr>
 *  All Rights Reserved
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307  USA
 */

#include <stdint.h>
#include <stdlib.h>
#include <string.h>

/* Tracker behaviour */
#define TRACKV1	1
#define TRACKV2	2
#define TRACKV3	4
#define TRACKV4	8
extern int track_mode;

/* Array handling */
#define array_append(array, item) \
    if((array) == NULL) \
    { \
        array = malloc(sizeof(array[0])); \
        array[0] = item; \
        array##_count = 1; \
    } \
    else \
    { \
        unsigned int count = array##_count; \
        while(!(count & 1)) \
            count >>= 1; \
        if(count == 1) \
            array = realloc(array, 2 * array##_count * sizeof(array[0])); \
        array[array##_count] = item; \
        array##_count++; \
    }

#define array_insert(array, item, pos) \
    if((array) == NULL) \
    { \
        array = malloc(sizeof(array[0])); \
        array[0] = item; \
        array##_count = 1; \
    } \
    else \
    { \
        unsigned int count = array##_count; \
        while(!(count & 1)) \
            count >>= 1; \
        if(count == 1) \
            array = realloc(array, 2 * array##_count * sizeof(array[0])); \
        memmove(array + pos + 1, array + pos, array##_count - pos); \
        array[pos] = item; \
        array##_count++; \
    }

#define array_del(array, pos) \
    if((array##_count) == 1) \
    { \
        free(array); \
        array = NULL; \
        array##_count = 0; \
    } \
    else \
    { \
        unsigned int count = array##_count - 1; \
        memmove(array + pos, array + pos + 1, array##_count - pos); \
        while(!(count & 1)) \
            count >>= 1; \
        if(count == 1) \
            array = realloc(array, (array##_count - 1) * sizeof(array[0])); \
        array##_count--; \
    }

/* Primitive structs */
typedef struct _pt2d
{
    double x, y;
}
pt2d_t;

typedef struct _pt3d
{
    double x, y, z;
}
pt3d_t;

/* Higher level structs */
typedef struct _ptsing
{
    pt2d_t point;
    double score;
}
ptsing_t;

typedef struct _scanline
{
    pt2d_t a;
    pt2d_t b;
}
scanline_t;

#define IMAGE_NORMAL 1
#define IMAGE_OPENCV 2

typedef struct _image
{
    unsigned int width, height, pitch;
    unsigned int depth;
    uint8_t * pixels;

    int _type;
    void * _private;
}
image_t;

typedef struct _projection
{
    ptsing_t * ptsing;
    unsigned int ptsing_count;

    scanline_t * scanline; 
    unsigned int scanline_count;

    pt2d_t intersect;
    pt3d_t vx, vy, vz, O, Os, I;
    double d, D, focal;

    double top, bottom, left, right;
}
projection_t;

#define LEFT 0
#define RIGHT 1

typedef struct _pair
{
    image_t image[2];
    projection_t projection[2];
    image_t image_norm[2];

    image_t depthmap_norm[2];
    image_t depthmap[2];

    double adjust;
}
pair_t;

/* Fast pixel stuff */
static inline int getpixel_fast(image_t *image, int x, int y, int *r, int *g, int *b)
{
    if(x < 0 || y < 0 || x >= image->width || y >= image->height)
    {
        *r = 255;
        *g = 0;
        *b = 255;
        return -1;
    }

    *b = image->pixels[y * image->pitch + x * image->depth];
    *g = image->pixels[y * image->pitch + x * image->depth + 1];
    *r = image->pixels[y * image->pitch + x * image->depth + 2];

    return 0;
}

static inline int setpixel_fast(image_t *image, int x, int y, int r, int g, int b)
{
    if(x < 0 || y < 0 || x >= image->width || y >= image->height)
        return -1;

    image->pixels[y * image->pitch + x * image->depth] = b;
    image->pixels[y * image->pitch + x * image->depth + 1] = g;
    image->pixels[y * image->pitch + x * image->depth + 2] = r;

    return 0;
}

