/*
 *  stereo -- Stereo matching software
 *  $Id: scanlines.c 162 2004-07-05 12:49:39Z eric $
 *  Copyright (c) 2004 Sam Hocevar <sam@zoy.org>
 *                     Éric Beets <ericbeets@free.fr>
 *  All Rights Reserved
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307  USA
 */

#include <stdint.h>
#include <stdio.h>

#include <cv.h>

#include "defs.h"
#include "main.h"
#include "geometry.h"

static int find_scanlines(image_t *left,
                          projection_t *left_data,
                          image_t *right,
                          projection_t *right_data);

static int intersect_scanlines(image_t *image,
                               projection_t *data);

/* Compute scanlines from the singular points */
int scanlines(image_t *left, projection_t *left_data, image_t *right, projection_t *right_data)
{
    int i, ret;

    /* Find scanlines */
    ret = find_scanlines(left, left_data, right, right_data);
    if(ret)
        return -1;

#if 0
//// REMOVE THAT CODE
    for(i = 0; i < left_data->scanline_count; i+=20)
    { CvPoint X, Y;
      X.x = left_data->scanline[i].a.x * 1;
      X.y = left_data->scanline[i].a.y * 1;
      Y.x = left_data->scanline[i].b.x * 1;
      Y.y = left_data->scanline[i].b.y * 1;
      cvLine(left->_private, X, Y, 0, 2.0, 0);
      X.x = right_data->scanline[i].a.x;
      X.y = right_data->scanline[i].a.y;
      Y.x = right_data->scanline[i].b.x;
      Y.y = right_data->scanline[i].b.y;
      cvLine(right->_private, X, Y, 0, 2.0, 0); }
//// ----------------
#endif
    /* Scanlines intersection */
    ret = intersect_scanlines(left, left_data);
    if(ret)
        return -1;

    ret = intersect_scanlines(right, right_data);
    if(ret)
        return -1;

    return 0;
}

static int find_scanlines(image_t *left,
                          projection_t *left_data,
                          image_t *right,
                          projection_t *right_data)
{
    int i, count, ret, size;
    int *line1,*line2,*lens1,*lens2;
    CvMat *cPoint1, *cPoint2, *mat, *status;
    CvMatrix3 *matrix;

    size = left_data->ptsing_count;

    /* Find the fundamental matrix */
    for(i = 0, count = 0; i < size; i++)
    {
        if(left_data->ptsing[i].score >= 1)
            count++;
    }

    printf("%d points trouves\n", count);

    mat = cvCreateMat(3, 3, CV_32F);
    cPoint1 = cvCreateMat(2, count, CV_32F);
    cPoint2 = cvCreateMat(2, count, CV_32F);
    status = cvCreateMat(1, count, CV_32F);

    for(i = 0, count = 0; i < size; i++)
    {
        if(left_data->ptsing[i].score >= 1)
        {
            cvmSet(cPoint1, 0, count, left_data->ptsing[i].point.x);
            cvmSet(cPoint1, 1, count, left_data->ptsing[i].point.y);
            cvmSet(cPoint2, 0, count, right_data->ptsing[i].point.x);
            cvmSet(cPoint2, 1, count, right_data->ptsing[i].point.y);
            count++;
        }
    }

    ret = cvFindFundamentalMat(cPoint1, cPoint2, mat, CV_FM_RANSAC,
                               1.0, 0.99, status);

    if(!ret)
        return -1;

    matrix = malloc(sizeof(CvMatrix3));
    for(i = 0; i < 9; i++)
        matrix->m[i / 3][i % 3] = (float)cvmGet(mat, i / 3, i % 3);

    /* Build the scanlines */
    cvMakeScanlines(matrix, cvSize(left->width, left->height),
                    0, 0, 0, 0, &count);
    line1 = calloc(count, sizeof(int) * 4);
    line2 = calloc(count, sizeof(int) * 4);
    lens1 = calloc(count * 2, sizeof(int) * 4);
    lens2 = calloc(count * 2, sizeof(int) * 4);
    cvMakeScanlines(matrix, cvSize(left->width, left->height),
                    line1, line2, lens1, lens2, &count);

    left_data->scanline = NULL;
    right_data->scanline = NULL;

    for(i = 0; i < count; i++)
    {
        scanline_t scanline;

        scanline.a.x = line1[i * 4];
        scanline.a.y = line1[i * 4 + 1];
        scanline.b.x = line1[i * 4 + 2];
        scanline.b.y = line1[i * 4 + 3];
        array_append(left_data->scanline, scanline);

        scanline.a.x = line2[i * 4];
        scanline.a.y = line2[i * 4 + 1];
        scanline.b.x = line2[i * 4 + 2];
        scanline.b.y = line2[i * 4 + 3];
        array_append(right_data->scanline, scanline);
    }

    printf("%d scanlines trouvees\n", count);

    return 0;
}

static int intersect_scanlines(image_t *image,
                               projection_t *data)
{
    int i, delta;
    pt2d_t *inter = NULL;
    unsigned int inter_count = 0;

    delta = data->scanline_count / 2;

    for(i = 0; i < data->scanline_count - delta; i++)
    {
        pt2d_t a = data->scanline[i].a;
        pt2d_t b = data->scanline[i].b;
        pt2d_t c = data->scanline[i + delta].a;
        pt2d_t d = data->scanline[i + delta].b;
        pt2d_t x;

        /* Trop petite */
        if(distance(a, b) < image->height || distance(a, b) < image->height)
            continue;

        x = intersection(a, b, c, d);

        array_append(inter, x);
    }

    data->intersect = barycentre(inter, inter_count);

    free(inter);

    return 0;
}

